/***********************************************/
/** Gestion des filtres de capture de packets **/
/***********************************************/

/** Fichiers d'inclusion **/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <linux/types.h>
#include <linux/filter.h>

/** Constantes **/

#define MLCOND_ANY	0
#define MLCOND_UHOST	1
#define MLCOND_NUHST	2
#define MLCOND_BRDCS	4
#define MLCOND_NBRDC	8

/** Structures globales **/

/* Definition d'une instruction BPF */

typedef struct {
  char *label;			/* Eventuel label de l'instruction */
  char condition;		/* Condition pour la compil. (MLCOND_XXXX) */
  int code;			/* Code de l'instruction */
  char *trueLabel;		/* Deplacement si vrai */
  char *falseLabel;		/* Deplacement si faux */
  char *argument;		/* Champ multi-usage */
  } machineLanguage;

/* Structure de stockage des valeurs des labels */

typedef struct {
  char *name;			/* Nom du label */
  int value;			/* Numero de l'instruction associee */
  } labelTable;

/* Structure de stockage des valeurs des variables */

typedef struct {
  char *name;			/* Nom de la variable */
  long int value;		/* Valeur de la variable */
  } variableTable;

/** Variables globales **/

machineLanguage filterProgram[]={
  {NULL,      MLCOND_ANY,
	      BPF_LD|BPF_H|BPF_ABS, NULL, NULL, "12"},
  {NULL,      MLCOND_ANY,
	      BPF_JMP|BPF_JEQ,      "srcaddr", "ipv6", "0x800"},
  {"ipv6",    MLCOND_ANY,
	      BPF_JMP|BPF_JEQ,      "srcaddr", "arp", "0x86dd"},
  {"arp",     MLCOND_ANY,
	      BPF_JMP|BPF_JEQ,      "srcaddr", "rarp", "0x806"},
  {"rarp",    MLCOND_ANY,
	      BPF_JMP|BPF_JEQ,      "srcaddr", "dummy", "0x8035"},
  {"dummy",   MLCOND_ANY,
	      BPF_JMP|BPF_JEQ,      "srcaddr", "fail", "0x1111"},
  {"srcaddr", MLCOND_UHOST,
	      BPF_LD|BPF_W|BPF_ABS, NULL, NULL, "8"},
  {NULL,      MLCOND_UHOST,
	      BPF_JMP|BPF_JEQ,      "next2", "dstaddr", "MacLow"},
  {"next2",   MLCOND_UHOST,
	      BPF_LD|BPF_H|BPF_ABS, NULL, NULL, "6"},
  {NULL,      MLCOND_UHOST,
	      BPF_JMP|BPF_JEQ,      "brdcast", "dstaddr", "MacHi"},
  {"dstaddr", MLCOND_UHOST,
	      BPF_LD|BPF_W|BPF_ABS, NULL, NULL, "2"},
  {NULL,      MLCOND_UHOST,
	      BPF_JMP|BPF_JEQ,      "next3", "fail", "MacLow"},
  {"next3",   MLCOND_UHOST,
	      BPF_LD|BPF_H|BPF_ABS, NULL, NULL, "0"},
  {NULL,      MLCOND_UHOST,
	      BPF_JMP|BPF_JEQ,      "brdcast", "fail", "MacHi"},
  {"brdcast", MLCOND_NBRDC,
	      BPF_LD|BPF_W|BPF_ABS, NULL, NULL, "2"},
  {NULL,      MLCOND_NBRDC,
	      BPF_JMP|BPF_JEQ,      "next4", "success", "0xffffffff"},
  {"next4",   MLCOND_NBRDC,
	      BPF_LD|BPF_H|BPF_ABS, NULL, NULL, "0"},
  {NULL,      MLCOND_NBRDC,
	      BPF_JMP|BPF_JEQ,      "fail", "success", "0xffff"},
  {"success", MLCOND_ANY,
	      BPF_LD|BPF_LEN,       NULL, NULL, "0"},
  {NULL,      MLCOND_ANY,
	      BPF_RET|BPF_A,        NULL, NULL, "0"},
  {"fail",    MLCOND_ANY,
	      BPF_RET|BPF_K,        NULL, NULL, "0"},
  {NULL,-1,0,NULL,NULL,0} 
};

/** Fonctions de calcul de tailles **/

int instNb(machineLanguage *program,int conditions){
  int nb=0;
  while(program->condition>=0){
    if((program->condition&conditions)==program->condition) nb++;
    program++;
    }
  return nb;
  }

int labelNb(machineLanguage *program,int conditions){
  int nb=0;
  while(program->condition>=0){
    if(program->label!=NULL) nb++;
    program++;
    }
  return nb;
  }

/** Fonction de calcul des valeurs de labels **/

labelTable *computeLabels(machineLanguage *program,int conditions){
  int len=labelNb(program,conditions)+1;
  labelTable *result=(labelTable *)malloc(len*sizeof(labelTable));
  int index=0,nb=0;
  while(program->condition>=0){
    if(program->label!=NULL){
      result[index].name=program->label;
      result[index++].value=nb;
      }
    if((program->condition&conditions)==program->condition) nb++;
    program++;
    }
  result[index].name=NULL;
  return result;
  }

/** Trouve la valeur d'un label **/

int labelValue(labelTable *table,char *name){
  if(name==NULL) return -1;
  while(table!=NULL&&table->name!=NULL){
    if(strcmp(table->name,name)==0) return table->value;
    table++;
    }
  { int result=-1;
    if(sscanf(name,"0x%x",&result)!=1) sscanf(name,"%d",&result);
    return result;
    }
  }

/** Trouve la valeur d'une variable **/

long int variableValue(variableTable *table,char *name){
  if(name==NULL) return -1;
  while(table!=NULL&&table->name!=NULL){
    if(strcmp(table->name,name)==0) return table->value;
    table++;
    }
  { long int result=-1;
    if(sscanf(name,"0x%lx",&result)!=1) sscanf(name,"%ld",&result);
    return result;
    }
  }

/** Compile un programme **/

struct sock_fprog *compileProgram(
    machineLanguage *program,int conditions,variableTable *vtable){
  labelTable *ltable=computeLabels(program,conditions);
  struct sock_fprog *result;
  int index=0,len=instNb(program,conditions);

  result=(struct sock_fprog *)malloc(sizeof(struct sock_fprog));
  result->filter=(struct sock_filter *)malloc(len*sizeof(struct sock_filter));
  result->len=len;
  while(program->condition>=0){
    if((program->condition&conditions)==program->condition){
      long int value=variableValue(vtable,program->argument);
      int jt=labelValue(ltable,program->trueLabel);
      int jf=labelValue(ltable,program->falseLabel);
      result->filter[index].code=program->code;
      result->filter[index].jt=jt<0?0:jt-index-1;
      result->filter[index].jf=jf<0?0:jf-index-1;
      result->filter[index].k=value;
      index++;
      }
    program++;
    }
  
  return result;
  }

/** Interface avec le programme de lecture et d'ecriture de paquest Ethernet **/

struct sock_fprog *makeFilter(unsigned char *mac,unsigned char broadcast){
  int flags=((mac==NULL)?MLCOND_NUHST:MLCOND_UHOST)|
	    (broadcast?MLCOND_BRDCS:MLCOND_NBRDC);
  variableTable variables[3];
  if(mac!=NULL){
    variables[0].name="MacLow";
    variables[0].value=(((((mac[2]<<8)|mac[3])<<8)|mac[4])<<8)|mac[5];
    variables[1].name="MacHi";
    variables[1].value=mac[0]<<8|mac[1];
    variables[2].name=NULL;
    }
  else variables[0].name=NULL;
  return compileProgram(filterProgram,flags,variables);
  }

/** Affiche un programme compile **/

void displayProgram(struct sock_fprog *program){
  int i;
  printf("{\n");
  for(i=0;i<program->len;i++){
    struct sock_filter *block=program->filter+i;
    printf("/* %4d */  { 0x%x, %d, %d, 0x%x }%c\n",i,
           block->code,block->jt,block->jf,block->k,i<program->len-1?',':' ');
    }
  printf("  }\n");
  fflush(stdout);
  }
