/*********************************************/
/** Lecture et criture de paquets Ethernet **/
/*********************************************/

/** Fichiers d'inclusion **/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <signal.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <linux/filter.h>

#include "bpf.h"

/** Constantes **/

#define BUFFER_SIZE	2048
#define MAX_COLUMNS	16
#define ETHER_ADDR_LEN	6

#define MODE_DISPLAY	0
#define MODE_SEND	1

/** Variables globales **/

char *ifname="eth0";		/* Interface reseau a utiliser */
int mode=MODE_DISPLAY;		/* Mode d'utilisation (lire ou ecrire) */
unsigned char broadcast=1;	/* Capture ou non des diffusions */
unsigned char unicast=0;	/* Capture limitee a une adresse MAC ou non */
				/* adresse MAC a surveiller */
unsigned char macAddress[ETHER_ADDR_LEN];

/** Affichage de la syntaxe de la commande **/

void usage(char *name){
  fprintf(stderr,"Usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"  [[-i|--interface] <net interface>] (default %s)\n",ifname);
  fprintf(stderr,"  [-d|--display-mode] (default mode)\n");
  fprintf(stderr,"  [-s|--send-mode]\n");
  fprintf(stderr,"  [-b|--no-broadcast]\n");
  fprintf(stderr,"  [[-u|--unicast] <mac address>]");
  fprintf(stderr," (no restriction by default)\n");
  exit(-1);
  }

/** Recuperation des arguments **/

void parseOptions(int argc,char **argv)
{
  int index;
  int c;

  /* Definition des options */

  static char *options="i:dsbu:";
  static const struct option longOptions[]={
    {"interface",1,0,'i'},
    {"display-mode",0,0,'d'},
    {"send-mode",0,0,'s'},
    {"no-broadcast",0,0,'b'},
    {"unicast",1,0,'u'},
    {0,0,0,0}
    } ;

  /* Analyse lexicale */

  while((c=getopt_long(argc,argv,options,longOptions,&index))>0)
    switch(c){
      case 'i': ifname=optarg; break;
      case 'd': mode=MODE_DISPLAY; break;
      case 's': mode=MODE_SEND; break;
      case 'b': broadcast=0; break;
      case 'u':{
        int x[ETHER_ADDR_LEN];
	int i,nb=sscanf(optarg,"%x:%x:%x:%x:%x:%x",x+0,x+1,x+2,x+3,x+4,x+5);
	if(nb!=ETHER_ADDR_LEN) usage(argv[0]);
	for(i=0;i<ETHER_ADDR_LEN;i++) macAddress[i]=x[i];
	unicast=1;
        break; }
      default: usage(argv[0]); break;
      }
  if(optind<argc) usage(argv[0]);
}

/** Recherche du numero systeme d'une interface **/

int getInterfaceIndex(int sock,char *ifname){
  struct ifreq ifr;
  strcpy(ifr.ifr_name,ifname);
  if(ioctl(sock,SIOCGIFINDEX,&ifr)<0){
    perror("getInterfaceIndex.ioctl");
    exit(-1);
    }
  return ifr.ifr_ifindex;
  }

/** Passage d'une interface reseau en mode PROMISCUOUS **/

void setPromiscuousMode(int sock,char *ifname,unsigned char set){
  static short int flags;
  struct ifreq ifr;
  strcpy(ifr.ifr_name,ifname);
  if(ioctl(sock,SIOCGIFFLAGS,&ifr)<0){
    perror("setPromiscuousMode.ioctl(SIOCGIFFLAGS)");
    exit(-1);
    }
  if(set==0) ifr.ifr_flags=flags; 
  else{ flags=ifr.ifr_flags;  ifr.ifr_flags|=IFF_PROMISC; }
  if(ioctl(sock,SIOCSIFFLAGS,&ifr)<0){
    perror("setPromiscuousMode.ioctl(SIOCSIFFLAGS)");
    exit(-1);
    }
  }

/** Mise en place d'un filtre sur une interface reseau **/

void setFilter(int sock,char *ifname,struct sock_fprog *filter){
  setsockopt(sock,SOL_SOCKET,SO_ATTACH_FILTER,filter,sizeof(struct sock_fprog));
  }

/** Ouverture de la socket **/

int openRawEthernetSocket(char *ifname){
  struct sockaddr_ll address;
  int proto=htons(ETH_P_ALL);
  int result;

  if((result=socket(PF_PACKET,SOCK_RAW,proto))<0){
    perror("openRawEthernetSocket.socket");
    exit(1);
  }

  memset((void *)&address,0,sizeof(struct sockaddr_ll));
  address.sll_family=AF_PACKET;
  address.sll_protocol=proto;
  address.sll_ifindex=getInterfaceIndex(result,ifname);
  if((bind(result,(struct sockaddr *)&address,sizeof(struct sockaddr_ll)))<0){
    perror("openRawEthernetSocket.bind");
    exit(1);
  }

  return result;
  }

/** Adresse de la socket destination **/

void initTargetAddress(int sock,char *ifname,struct sockaddr_ll *address){
  memset((void *)address,0,sizeof(struct sockaddr_ll));
  address->sll_family=AF_PACKET;
  address->sll_halen=ETHER_ADDR_LEN;
  address->sll_ifindex=getInterfaceIndex(sock,ifname);
  }

/** Lecture d'un paquet sur la socket **/

void readPacket(int sock,char *packet,int *size){
  *size=recvfrom(sock,packet,*size,0,NULL,NULL);
  }

/** Ecriture d'un paquet sur la socket **/

void writePacket(int sock,struct sockaddr_ll *address,char *packet,int size){
  int asize=sizeof(struct sockaddr_ll); 
  memcpy(address->sll_addr,packet,ETHER_ADDR_LEN);
  sendto(sock,packet,size,0,(struct sockaddr *)address,asize);
  }

/** Affichage d'un paquet sur la sortie standard **/

void displayPacket(FILE *stream,char *packet,int size){
  int i,count=MAX_COLUMNS;
  for(i=0;i<size;i++){
    fprintf(stream,"%2.2x ",(unsigned char)packet[i]);
    count--;
    if(count<=0){ fprintf(stream,"\n"); count=MAX_COLUMNS; }
    }
  if(count!=MAX_COLUMNS) fprintf(stream,"\n");
  }

/** Lecture d'un paquet sur l'entree standard **/

void scanPacket(FILE *stream,char *packet,int *size){
  int max=*size;
  int byte;
  *size=0;
  while(fscanf(stream,"%x",&byte)==1 && *size<max){
    packet[*size]=(unsigned char)byte;
    (*size)++;
    }
  }

/** Fonction principale **/

int sock;

void clean(int signal){
  if(unicast) setPromiscuousMode(sock,ifname,0);
  exit(0);
  }

int main(int argc, char **argv){
  struct sockaddr_ll address;
  int size;
  char buffer[BUFFER_SIZE];

  parseOptions(argc,argv);
  sock=openRawEthernetSocket(ifname);
  if(mode==MODE_DISPLAY){
    struct sock_fprog *filter=makeFilter(unicast?macAddress:NULL,broadcast);
    displayProgram(filter);
    setFilter(sock,ifname,filter);
    if(unicast) setPromiscuousMode(sock,ifname,1);
    signal(SIGINT,clean);
    while(1){
      time_t t=time(NULL);
      size=BUFFER_SIZE;
      readPacket(sock,buffer,&size);
      if(size<=0) break;
      fprintf(stdout,"Packet received at %s",ctime((const time_t *)&t));
      displayPacket(stdout,buffer,size);
      fprintf(stdout,"\n");
      fflush(stdout);
      }
    }
  else{
    initTargetAddress(sock,ifname,&address);
    size=BUFFER_SIZE;
    scanPacket(stdin,buffer,&size);
    writePacket(sock,&address,buffer,size);
    }
  exit(0);
  }
